unit EListBox;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics,
  Controls, Forms, Dialogs, StdCtrls;

const
  MaxTabs = 40;

type
  TCompareItemsEvent = procedure(Control: TWinControl;
    const S1, S2: String; var Leg: Integer) of object;

  TExtListBox = class(TListBox)
  private
    FDlgTabs, FPixTabs: array[0..MaxTabs] of Integer;
    FHighTab: Integer;
    DlgBaseX : Integer;
    FOnCompare: TCompareItemsEvent;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure LBAddString(var Msg: TMessage); message LB_ADDSTRING;
    procedure LBInsertString(var Msg: TMessage); message LB_INSERTSTRING;
    procedure LBDeleteString(var Msg: TMessage); message LB_DELETESTRING;
    procedure LBResetContent(var Msg: TMessage); message LB_RESETCONTENT;
    procedure CMFontChanged(var Msg: TMessage); message CM_FONTCHANGED;
    function CompareItems(const S1, S2: String): Integer;
    function GetScrollWidth: Integer;
    procedure SetScrollWidth(Value: Integer);
    procedure SetOnCompare(Value: TCompareItemsEvent);
  protected
    procedure AllWidths;
    procedure NewWidth(P: PChar);
    function WidthOfString(const S: String): Integer;
    function WidthOfPChar(P: PChar): Integer;
  public
    procedure SpecialSort;
    procedure FindPartial(S: String);
    procedure FindExact(S: String);
  published
    procedure SetTabsDlg(const Tabs: array of Integer);
    procedure SetTabsPix(const Tabs: array of Integer);
    property ScrollWidth: Integer read GetScrollWidth write SetScrollWidth;
    property OnCompare: TCompareItemsEvent read FOnCompare write SetOnCompare;
  end;

procedure Register;

implementation

procedure TExtListBox.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params do
    Style := Style or LBS_USETABSTOPS;
  DlgBaseX := LoWord(GetDialogBaseUnits);
end;

function TExtListBox.GetScrollWidth: Integer;
begin
  Result := Perform(LB_GETHORIZONTALEXTENT, 0, 0);
end;

procedure TExtListBox.SetScrollWidth(Value : Integer);
begin
  Perform(LB_SETHORIZONTALEXTENT, Value, 0);
end;

procedure TExtListBox.LBAddString(var Msg: TMessage);
begin
  inherited;
  NewWidth(PChar(Msg.lParam));
end;

procedure TExtListBox.LBInsertString(var Msg: TMessage);
begin
  inherited;
  NewWidth(PChar(Msg.lParam));
end;

procedure TExtListBox.LBDeleteString(var Msg: TMessage);
begin
  inherited;
  AllWidths;
end;

procedure TExtListBox.LBResetContent(var Msg: TMessage);
begin
  inherited;
  ScrollWidth := 0;
end;

procedure TExtListBox.CMFontChanged(var Msg: TMessage);
begin
  inherited;
  AllWidths;
end;

function TExtListBox.WidthOfString(const S : String): Integer;
var
  Buffer: array[0..255] of Char;
begin
  StrPCopy(Buffer, S);
  Result := WidthOfPChar(Buffer);
end;

function TExtListBox.WidthOfPChar(P : PChar): Integer;
begin
  Canvas.Font := Font;
  Result := LoWord(GetTabbedTextExtent(Canvas.Handle, P, StrLen(P),
    fHighTab+1, fPixTabs)) + Canvas.TextWidth('X');
end;

procedure TExtListBox.AllWidths;
var
  N, NewWid, Wid : Integer;
begin
  NewWid := 0;
  for N := 0 to Items.Count-1 do begin
    Wid := WidthOfString(Items[N]);
    if Wid > NewWid then NewWid := Wid;
  end;
  ScrollWidth := NewWid;
end;

procedure TExtListBox.NewWidth(P : PChar);
var
  Wid : Integer;
begin
  Canvas.Font := Font;
  Wid := WidthOfPChar(P);
  if Wid > ScrollWidth then ScrollWidth := Wid;
end;

procedure TExtListBox.SetTabsDlg(const Tabs: array of Integer);
var
  N : Integer;
begin
  FHighTab := High(Tabs);
  if FHighTab > MaxTabs then FHighTab := MaxTabs;
  for N := 0 to fHighTab do begin
    FDlgTabs[N] := Tabs[N];
    FPixTabs[N] := Round((Tabs[N] * DlgBaseX) / 4);
  end;
  Perform(LB_SETTABSTOPS, FHighTab+1, LongInt(@FDlgTabs));
  Refresh;
  AllWidths;
end;

procedure TExtListBox.SetTabsPix(const Tabs: array of Integer);
var
  N : Word;
begin
  FHighTab := High(Tabs);
  if FHighTab > MaxTabs then FHighTab := MaxTabs;
  for N := 0 to FHighTab do begin
    FDlgTabs[N] := Round((Tabs[N]*4) / DlgBaseX);
    FPixTabs[N] := Tabs[N];
  end;
  Perform(LB_SETTABSTOPS, FHighTab+1, LongInt(@FDlgTabs));
  Refresh;
  AllWidths;
end;

function TExtListBox.CompareItems(const S1, S2: String): Integer;
begin
  if Assigned(FOnCompare) then
    FOnCompare(Self, S1, S2, Result) else
    Result := AnsiCompareStr(S1, S2);
end;

procedure TExtListBox.SpecialSort;
var
  List: TStringList;
  N: Integer;
  function FindPlace(const S: String): Integer;
  var
    vLo, vHi, vIdx: Integer;
  begin
    vLo := 0;
    vHi := List.Count-1;
    while vLo <= vHi do begin
      VIdx := (vLo + vHi) shr 1;
      if CompareItems(List[vIdx], S) < 0 then
        vLo := vIdx + 1 else
        vHi := vIdx - 1;
    end;
    Result := vLo;
  end;
begin
  Sorted := False;
  List := TStringList.Create;
  try
    for N := 0 to Items.Count-1 do
      List.InsertObject(FindPlace(Items[N]), Items[N], Items.Objects[N]);
    Items.Assign(List);
  finally
    List.Free;
  end;
end;

procedure TExtListBox.SetOnCompare(Value: TCompareItemsEvent);
begin
  FOnCompare := Value;
  SpecialSort;
end;

procedure TExtListBox.FindPartial(S: String);
var
  Buf: array[0..255] of Char;
begin
  StrPCopy(Buf, S);
  ItemIndex := Perform(LB_SELECTSTRING, 0, LongInt(@Buf));
end;

procedure TExtListBox.FindExact(S: String);
var
  Buf: array[0..255] of Char;
begin
  StrPCopy(Buf, S);
  ItemIndex := Perform(LB_FINDSTRINGEXACT, 0, LongInt(@Buf));
end;

procedure Register;
begin
  RegisterComponents('Samples', [TExtListBox]);
end;

end.
