unit SplitBar;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls;

type
  TSplitBar = class(TCustomControl)
  private
    FForm: TForm;
    FSplit: TPoint;
    FSizing: Boolean;
    FKind: TScrollBarKind;
    FSizeControl: TControl;
    FMaxSize: Integer;
    FMinSize: Integer;
    function Length: Integer;
    function Breadth: Integer;
    procedure DrawSizingLine;
    procedure ResizeControl;
    procedure SetKind(Value: TScrollBarKind);
  protected
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation);
      override;
    procedure BeginSizing;
    procedure ChangeSizing(X, Y: Integer);
    procedure EndSizing;
    property Sizing: Boolean read FSizing;
  published
    property Align default alLeft;
    property Color default clBtnFace;
    property Cursor default crHSplit;
    property DragCursor default crHSplit;
    property Enabled;
    property Height default 50;
    property Kind: TScrollBarKind read FKind write SetKind default sbVertical;
    property ParentColor;
    property ParentShowHint;
    property MaxSize: Integer read FMaxSize write FMaxSize;
    property MinSize: Integer read FMinSize write FMinSize;
    property SizeControl: TControl read FSizeControl write FSizeControl;
    property ShowHint;
    property Visible;
    property Width default 2;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Samples', [TSplitBar]);
end;

function CToC(C1, C2: TControl; P: TPoint): TPoint;
begin
  Result := C1.ScreenToClient(C2.ClientToScreen(P));
end;

constructor TSplitBar.Create(AOwner: TComponent);
  function GetForm(C: TComponent): TForm;
  begin
    if C is TForm then
      Result := TForm(C) else
      Result := GetForm(C.Owner);
  end;
begin
  inherited Create(AOwner);
  FSizing := False;
  FForm  := GetForm(Owner);
  FKind  := sbVertical;
  Width  := 2;
  Height := 50;
  Align  := alLeft;
  Color  := clBtnFace;
  Cursor := crHSplit;
  DragCursor := crHSplit;
end;

procedure TSplitBar.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = FSizeControl) then
    FSizeControl := nil;
end;

function TSplitBar.Length: Integer;
begin
  if FKind = sbVertical then
    Result := Height else
    Result := Width;
end;

function TSplitBar.Breadth: Integer;
begin
  if FKind = sbVertical then
    Result := Width else
    Result := Height;
end;

procedure TSplitBar.DrawSizingLine;
var
  p: TPoint;
begin
  p := CToC(FForm, Self, FSplit);
  with FForm.Canvas do begin
    MoveTo(p.X, p.Y);
    if FKind = sbVertical then
      LineTo(p.X, CToC(FForm, Self, Point(0, Length)).Y) else
      LineTo(CToC(FForm, Self, Point(Length, 0)).X, p.Y);
  end;
end;

procedure TSplitBar.BeginSizing;
begin
  SetCaptureControl(Self);
  FSizing := True;
  if FKind = sbVertical then
    FSplit := Point(0, Top) else
    FSplit := Point(Left, 0);
  with FForm.Canvas do begin
    Handle := GetDCEx(FForm.Handle, 0,
      DCX_CACHE or DCX_CLIPSIBLINGS or DCX_LOCKWINDOWUPDATE);
    Pen.Color := clWhite;
    Pen.Width := Breadth;
    Pen.Mode  := pmXOR;
  end;
  DrawSizingLine;
end;

procedure TSplitBar.ChangeSizing(X, Y: Integer);
begin
  DrawSizingLine;
  if FKind = sbVertical then
    FSplit.X := X else
    FSplit.Y := Y;
  DrawSizingLine;
end;

procedure TSplitBar.EndSizing;
var
  dc: HDC;
begin
  DrawSizingLine;
  SetCaptureControl(nil);
  FSizing := False;
  with FForm do begin
    dc := Canvas.Handle;
    Canvas.Handle := 0;
    ReleaseDC(Handle, dc);
  end;
  if SizeControl <> nil then
    ResizeControl;
end;

procedure TSplitBar.ResizeControl;
  function GetSize(Size: Integer): Integer;
  begin
    if Size < FMinSize then
      Result := FMinSize
    else if (FMaxSize > 0) and (Size > FMaxSize) then
      Result := FMaxSize
    else
      Result := Size;
  end;
var
  p: TPoint;
begin
  p := CToC(FSizeControl, Self, FSplit);
  with FSizeControl do
    case Align of
      alLeft:   Width  := GetSize(p.X);
      alRight:  Width  := GetSize(Width - p.X);
      alTop:    Height := GetSize(p.Y);
      alBottom: Height := GetSize(Height - p.Y);
    end;
end;

procedure TSplitBar.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseDown(Button, Shift, X, Y);
  if (Button = mbLeft) and (Shift = [ssLeft]) then
    BeginSizing;
end;

procedure TSplitBar.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseMove(Shift, X, Y);
  if FSizing then ChangeSizing(X, Y);
end;

procedure TSplitBar.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseUp(Button, Shift, X, Y);
  if FSizing then EndSizing;
end;

procedure TSplitBar.SetKind(Value: TScrollBarKind);
  procedure Swap;
  var
    i: Integer;
  begin
    i := Width;
    Width := Height;
    Height := i;
  end;
begin
  FKind := Value;
  if FKind = sbVertical then begin
    if not (Align in [alLeft, alRight]) then
      Align := alLeft;
    Cursor := crHSplit;
    DragCursor := crHSplit;
    if Width > Height then Swap;
  end
  else begin  {FKind = sbHorizontal}
    if not (Align in [alTop, alBottom]) then
      Align := alTop;
    Cursor := crVSplit;
    DragCursor := crVSplit;
    if Height > Width then Swap;
  end;
end;

end.
